@@ -29,106 +29,97 @@ Usage:
     oracle.consensus()
 
 """
-from __future__ import division, print_function, unicode_literals, absolute_import
+from __future__ import division, absolute_import
 import sys
-try:
-    import cdecimal
-    sys.modules["decimal"] = cdecimal
-except:
-    pass
 import os
 import getopt
-from decimal import Decimal, getcontext, ROUND_HALF_EVEN
 import numpy as np
 import numpy.ma as ma
 from weightedstats import weighted_median
+from six.moves import xrange as range
 
 __title__      = "pyconsensus"
-__version__    = "0.2"
+__version__    = "0.2.1"
 __author__     = "Paul Sztorc and Jack Peterson"
 __license__    = "GPL"
 __maintainer__ = "Jack Peterson"
 __email__      = "jack@tinybike.net"
 
-# Python 3 compatibility
-from six.moves import xrange as range
-_IS_PYTHON_3 = sys.version_info[0] == 3
-identity = lambda x : x
-if _IS_PYTHON_3:
-    u = identity
-else:
-    import codecs
-    def u(string):
-        return codecs.unicode_escape_decode(string)[0]
-
-getcontext().rounding = ROUND_HALF_EVEN
-
 class Oracle(object):
 
-    def __init__(self, votes=None, decision_bounds=None, weights=-1,
-                 catch_p=.1, max_row=5000, verbose=False):
+    def __init__(self, votes=None, decision_bounds=None, reputation=None,
+                 catch_tolerance=0.1, max_row=5000, alpha=0.1, verbose=False):
+        """
+        Args:
+          votes (list-of-lists): votes matrix; rows = voters, columns = Decisions.
+          Scales (list): list of dicts for each Decision
+            {
+              scaled (bool): True if scalar, False if binary (boolean)
+              min (float): minimum allowed value (0 if binary)
+              max (float): maximum allowed value (1 if binary)
+            }
+
+        """
         self.votes = ma.masked_array(votes, np.isnan(votes))
         self.decision_bounds = decision_bounds
-        self.weights = weights
-        self.catch_p = catch_p
+        self.catch_tolerance = catch_tolerance
         self.max_row = max_row
+        self.alpha = alpha
         self.verbose = verbose
+        self.num_votes = len(votes)
+        if reputation is None:
+            self.total_rep = self.num_votes
+            self.reputation = np.array([[1 / float(self.num_votes)]] * self.num_votes)
+            self.rep_coins = (np.copy(self.reputation) * 10**6).astype(int)
+        else:
+            self.total_rep = sum(np.array(reputation).flatten())
+            self.reputation = np.array([i / float(self.total_rep) for i in reputation])
+            self.rep_coins = np.array([map(int, i) for i in np.abs(reputation) * 10**6])
 
-    def Rescale(self, UnscaledMatrix, Scales):
+    def Rescale(self):
         """Forces a matrix of raw (user-supplied) information
         (for example, # of House Seats, or DJIA) to conform to
-        svd-appropriate range.
+        SVD-appropriate range.
 
         Practically, this is done by subtracting min and dividing by
         scaled-range (which itself is max-min).
 
         """
-        # Calulate multiplicative factors   
+        # Calulate multiplicative factors
         InvSpan = []
-        for scale in Scales:
+        for scale in self.decision_bounds:
             InvSpan.append(1 / float(scale["max"] - scale["min"]))
 
         # Recenter
-        OutMatrix = ma.copy(UnscaledMatrix)
-        cols = UnscaledMatrix.shape[1]
+        OutMatrix = ma.copy(self.votes)
+        cols = self.votes.shape[1]
         for i in range(cols):
-            OutMatrix[:,i] -= Scales[i]["min"]
+            OutMatrix[:,i] -= self.decision_bounds[i]["min"]
 
         # Rescale
         NaIndex = np.isnan(OutMatrix)
         OutMatrix[NaIndex] = 0
-
         OutMatrix = np.dot(OutMatrix, np.diag(InvSpan))
-        OutMatrix[NaIndex] = np.nan
 
         return OutMatrix
 
-    def MeanNa(self, Vec):
+    def MeanNa(self, v):
         """Takes masked array, replaces missing values with array mean."""
-        MM = np.mean(Vec)
-        Vec[np.where(Vec.mask)] = MM
-        return Vec
+        v[np.where(v.mask)] = np.mean(v)
+        return v
         
-    def GetWeight(self, Vec, AddMean=0):
+    def GetWeight(self, v):
         """Takes an array, and returns proportional distance from zero."""
-        New = abs(Vec)
+        v = abs(v)
+        if np.sum(v) == 0:
+            v += 1
+        return v / np.sum(v)
 
-        # Add the mean to each element of the vector
-        if AddMean == 1:
-            New += np.mean(New)
-
-        # Catch an error here
-        if np.sum(New) == 0:
-            New += 1
-
-        # Normalize
-        return New / np.sum(New)
-
-    def Catch(self, X, Tolerance=0):
+    def Catch(self, X):
         """Forces continuous values into bins at 0, .5, and 1"""
-        if X < (.5-(Tolerance/2)):
+        if X < 0.5 * (1 - self.catch_tolerance):
             return 0
-        elif X > (.5+(Tolerance/2)):
+        elif X > 0.5 * (1 + self.catch_tolerance):
             return 1
         else:
             return .5
@@ -144,67 +135,65 @@ class Oracle(object):
             Out.append(Weight[i] / Expected[i])
         return Out
 
-    def ReWeight(self, Vec):
+    def ReWeight(self, v):
         """Get the relative influence of numbers, treat NaN as influence-less."""
-        Out = Vec
-        Exclude = np.isnan(Vec)
+        exclude = np.isnan(v)
 
         # Set missing values to 0
-        Out[Exclude] = 0
+        v[exclude] = 0
 
         # Normalize
-        return Out / np.sum(Out)
+        return v / np.sum(v)
 
-    def DemocracyCoin(self, votes):
-        """No coin distribution given, assuming democracy: one person, one vote."""
-        return self.GetWeight(np.array([[1]] * len(votes)))
+    def WeightedCov(self, votes_filled):
+        """Weights are the number of coins people start with, so the aim of this
+        weighting is to count 1 vote for each of their coins -- e.g., guy with 10
+        coins effectively gets 10 votes, guy with 1 coin gets 1 vote, etc.
 
-    def WeightedCov(self, votes, reputation=-1):
-        """Takes 1] a masked array, and 2] an [n x 1] dimentional array of
-        weights, and computes the weighted covariance matrix and center of
-        a given array.
+        Takes 1] a masked array, and 2] an [n x 1] dimentional array of weights,
+        and computes the weighted covariance matrix and center of a given array.
         Taken from
         http://stats.stackexchange.com/questions/61225/correct-equation-for-weighted-unbiased-sample-covariance
         """
-        if type(reputation) is int:
-            reputation = self.DemocracyCoin(votes)
+        # Compute the weighted mean (of all voters) for each decision
+        # e.g. decision_1_mean = (voter_0_decision_1 + voter_1_decision_1 + voter_2_decision_1 + ...) / N
+        weighted_mean = ma.average(votes_filled, axis=0, weights=self.rep_coins.squeeze())
 
-        Coins = np.array([map(int, i) for i in reputation * 10**6])
-
-        # Computing the weighted sample mean (fast, efficient and precise)
-        weighted_sample_mean = ma.average(votes, axis=0, weights=np.hstack(Coins))
-        XM = np.matrix(votes - weighted_sample_mean) # xm = X diff to mean
+        # Differences from the mean
+        mean_deviation = np.matrix(votes_filled - weighted_mean)
 
         # Compute the unbiased weighted sample covariance
-        sigma2 = np.matrix(1/(np.sum(Coins)-1) * ma.multiply(XM, Coins).T.dot(XM));
+        #
+        # TODO for uniform weights, covariance_matrix should be the same as
+        # cov(mean_deviation.T), but is in fact the same as
+        # cov(mean_deviation.T, bias=1) -- why is this?
+        covariance_matrix = 1/(sum(self.rep_coins)-1) * ma.multiply(mean_deviation, self.rep_coins).T.dot(mean_deviation)
 
-        return {'Cov': np.array(sigma2), 'Center': np.array(XM)}
+        return covariance_matrix, mean_deviation
 
-    def WeightedPrinComp(self, votes, reputation=-1):
-        """Takes a matrix and row-weights and manually computes the statistical procedure known as Principal Components Analysis (PCA)
-        This version of the procedure is so basic, that it can also be thought of as merely a singular-value decomposition on a weighted covariance matrix."""      
-        wCVM = self.WeightedCov(votes, reputation)
-        SVD = np.linalg.svd(wCVM['Cov'])
+    def WeightedPrinComp(self, votes_filled):
+        """Principal Component Analysis (PCA) on the votes matrix.
+
+        The votes matrix has voters as rows and decisions as columns.
+
+        """
+        covariance_matrix, mean_deviation = self.WeightedCov(votes_filled)
+        SVD = np.linalg.svd(covariance_matrix)
 
         # First loading
         L = SVD[0].T[0]
 
         # First score
-        S = np.dot(wCVM['Center'], SVD[0]).T[0]
+        S = np.dot(mean_deviation, SVD[0]).T[0]
 
         return L, S
 
-    def GetRewardWeights(self, votes, reputation=-1, Alpha=.1, Verbose=False):
+    def GetRewardWeights(self, votes_filled):
         """Calculates new reputations using a weighted
         Principal Component Analysis (PCA).
 
         """
-        if type(reputation) is int:
-            reputation = self.DemocracyCoin(votes)
-
-        Results = self.WeightedPrinComp(votes, reputation)
-        # Need the old reputation back for the rest of this.
-        reputation = self.GetWeight(reputation)
+        Results = self.WeightedPrinComp(votes_filled)
         # The first loading is designed to indicate which Decisions were more 'agreed-upon' than others.
         FirstLoading = Results[0]
         # The scores show loadings on consensus (to what extent does this observation represent consensus?)
@@ -217,9 +206,9 @@ class Oracle(object):
         #  I originally tried doing this using math but after multiple failures I chose this ad hoc way.
         Set1 = FirstScore + abs(min(FirstScore))
         Set2 = FirstScore - max(FirstScore)
-        Old = np.dot(reputation.T, votes)
-        New1 = np.dot(self.GetWeight(Set1), votes)
-        New2 = np.dot(self.GetWeight(Set2), votes)
+        Old = np.dot(self.rep_coins.T, votes_filled)
+        New1 = np.dot(self.GetWeight(Set1), votes_filled)
+        New2 = np.dot(self.GetWeight(Set2), votes_filled)
 
         # Difference in sum of squared errors. If > 0, then New1 had higher
         # errors (use New2); conversely if < 0, then use New1.
@@ -228,55 +217,57 @@ class Oracle(object):
             AdjPrinComp = Set1
         if RefInd > 0:
             AdjPrinComp = Set2
-
-        if Verbose:
-            print("")
-            print("Estimations using: Previous reputation, Option 1, Option 2")
-            print(np.vstack([Old, New1, New2]).T)
-            print("")
-            print("Previous period reputations, Option 1, Option 2, Selection")
-            print(np.vstack([reputation.T, Set1, Set2, AdjPrinComp]).T)
       
         #Declared here, filled below (unless there was a perfect consensus).
-        RowRewardWeighted = reputation # (set this to uniform if you want a passive diffusion toward equality when people cooperate [not sure why you would]). Instead diffuses towards previous reputation (Smoothing does this anyway).
+        RowRewardWeighted = self.reputation # (set this to uniform if you want a passive diffusion toward equality when people cooperate [not sure why you would]). Instead diffuses towards previous reputation (Smoothing does this anyway).
         if max(abs(AdjPrinComp)) != 0:
             # Overwrite the inital declaration IFF there wasn't perfect consensus.
-            RowRewardWeighted = self.GetWeight(AdjPrinComp * (reputation / np.mean(reputation)).T)
+            RowRewardWeighted = self.GetWeight(AdjPrinComp * (self.reputation / np.mean(self.reputation)).T)
 
         #note: reputation/mean(reputation) is a correction ensuring Reputation is additive. Therefore, nothing can be gained by splitting/combining Reputation into single/multiple accounts.
               
         # Freshly-Calculated Reward (Reputation) - Exponential Smoothing
         # New Reward: RowRewardWeighted
         # Old Reward: reputation
-        SmoothedR = Alpha*RowRewardWeighted + (1-Alpha)*reputation.T
+        SmoothedR = self.alpha*RowRewardWeighted + (1-self.alpha)*self.reputation.T
 
         return {
             "FirstL": FirstLoading,
-            "OldRep": reputation.T,
+            "OldRep": self.reputation.T,
             "ThisRep": RowRewardWeighted,
-            "SmoothRep":SmoothedR,
+            "SmoothRep": SmoothedR,
         }
 
-    def GetDecisionOutcomes(self, Mtemp, ScaledIndex, Rep=-1, Verbose=False):
-        """Determines the Outcomes of Decisions based on the provided reputation (weighted vote)."""
+    def GetDecisionOutcomes(self, votes, ScaledIndex):
+        """Determines the Outcomes of Decisions based on the provided
+        reputation (weighted vote).
 
-        if type(Rep) is int:
-            Rep = self.DemocracyCoin(Mtemp)
-
-        RewardWeightsNA = Rep
+        """
         DecisionOutcomes_Raw = []
+        
+        # Iterate over columns
+        # import json
+        # print(json.dumps(self.decision_bounds, indent=3, sort_keys=True))
+        for i in range(votes.shape[1]):
 
-        # For each column:
-        for i in range(Mtemp.shape[1]):
-
-            # The Reputation of the row-players who DID provide
+            # The Reputation of the rows (players) who DID provide
             # judgements, rescaled to sum to 1.
-            Row = self.ReWeight(RewardWeightsNA[-Mtemp[..., i].mask])
+            # print(-votes[:,i])
+            # print(-votes[:,i].mask)
+            # print(self.reputation[-votes[:,i].mask])
+
+            # exclude = np.isnan(v)            
+            # # Set missing values to 0
+            # v[exclude] = 0
+            # # Normalize
+            # v / np.sum(v)
+
+            Row = self.ReWeight(self.reputation[-votes[:,i].mask])
 
             # The relevant Decision with NAs removed.
             # ("What these row-players had to say about the Decisions
             # they DID judge.")
-            Col = Mtemp[-Mtemp[..., i].mask, i]
+            Col = votes[-votes[:,i].mask, i]
 
             # Discriminate Based on Contract Type
             if not ScaledIndex[i]:
@@ -289,23 +280,21 @@ class Oracle(object):
 
         return np.array(DecisionOutcomes_Raw).T
 
-    def FillNa(self, votes_na, ScaledIndex, reputation=-1, CatchP=.1, Verbose=False):
+    def FillNa(self, votes_na, ScaledIndex):
         """Uses exisiting data and reputations to fill missing observations.
         Essentially a weighted average using all availiable non-NA data.
         How much should slackers who arent voting suffer? I decided this would
         depend on the global percentage of slacking.
         """
-        if type(reputation) is int:
-            reputation = self.DemocracyCoin(votes_na)
-
         # In case no Missing values, Mnew and votes_na will be the same.
         votes_na_new = ma.copy(votes_na)
 
         # Of course, only do this process if there ARE missing values.
         if votes_na.mask.any():
+
             # Our best guess for the Decision state (FALSE=0, Ambiguous=.5, TRUE=1)
             # so far (ie, using the present, non-missing, values).
-            DecisionOutcomes_Raw = self.GetDecisionOutcomes(votes_na, ScaledIndex, reputation, Verbose).squeeze()
+            DecisionOutcomes_Raw = self.GetDecisionOutcomes(votes_na, ScaledIndex).squeeze()
 
             # Fill in the predictions to the original M
             NAmat = votes_na.mask  # Defines the slice of the matrix which needs to be edited.
@@ -313,7 +302,7 @@ class Oracle(object):
 
             # Slightly complicated:
             NAsToFill = np.dot(NAmat, np.diag(DecisionOutcomes_Raw))
-            
+
             # This builds a matrix whose columns j:
             #   NAmat was false (the observation wasn't missing) - have a value of Zero
             #   NAmat was true (the observation was missing)     - have a value of the jth element of DecisionOutcomes.Raw (the 'current best guess')
@@ -326,52 +315,32 @@ class Oracle(object):
             for i in range(rows):
                 for j in range(cols):
                     if not ScaledIndex[j]:
-                        votes_na_new[i][j] = self.Catch(votes_na_new[i][j], CatchP)
+                        votes_na_new[i][j] = self.Catch(votes_na_new[i][j])
 
         return votes_na_new
 
     def consensus(self):
-        """
-        Args:
-          M0 (ma.masked_array): votes matrix; rows = voters, columns = Decisions.
-          Scales (list): list of dicts for each Decision
-            {
-              scaled (bool): True if scalar, False if binary (boolean)
-              min (float): minimum allowed value (0 if binary)
-              max (float): maximum allowed value (1 if binary)
-            }
+        """PCA-based consensus algorithm.
 
         Returns:
           dict: consensus results
 
         """
-        M0 = self.votes
-        Scales = self.decision_bounds
-        Rep = self.weights
-        CatchP = self.catch_p
-        MaxRow = self.max_row
-        Verbose = self.verbose
-
-        # Fill the default reputations (egalitarian) if none are provided...
-        # unrealistic and only for testing.
-        if type(Rep) is int:
-            Rep = self.DemocracyCoin(M0)
-
         # Fill the default scales (binary) if none are provided.
         # In practice, this would also never be used.
-        if Scales is None:
-            ScaledIndex = [False] * M0.shape[1]
-            MScaled = M0
+        if self.decision_bounds is None:
+            ScaledIndex = [False] * self.votes.shape[1]
+            MScaled = self.votes
         else:
-            ScaledIndex = [scale["scaled"] for scale in Scales]
-            MScaled = self.Rescale(M0, Scales)
+            ScaledIndex = [scale["scaled"] for scale in self.decision_bounds]
+            MScaled = self.Rescale()
 
         # Handle Missing Values
-        Filled = self.FillNa(MScaled, ScaledIndex, Rep, CatchP, Verbose)
+        votes_filled = self.FillNa(MScaled, ScaledIndex)
 
         # Consensus - Row Players
         # New Consensus Reward
-        PlayerInfo = self.GetRewardWeights(Filled, Rep, .1, Verbose)
+        PlayerInfo = self.GetRewardWeights(votes_filled)
         AdjLoadings = PlayerInfo['FirstL']
 
         # Column Players (The Decision Creators)
@@ -379,19 +348,18 @@ class Oracle(object):
         # Consensus - "Who won?" Decision Outcome    
         # Simple matrix multiplication ... highest information density at RowBonus,
         # but need DecisionOutcomes.Raw to get to that
-        DecisionOutcomes_Raw = np.dot(PlayerInfo['SmoothRep'], Filled).squeeze()
+        DecisionOutcomes_Raw = np.dot(PlayerInfo['SmoothRep'], votes_filled).squeeze()
 
         # Discriminate Based on Contract Type
-        ncols = Filled.shape[1]
-        for i in range(ncols):
+        for i in range(votes_filled.shape[1]):
             # Our Current best-guess for this Scaled Decision (weighted median)
             if ScaledIndex[i]:
-                DecisionOutcomes_Raw[i] = weighted_median(Filled[:,i],
+                DecisionOutcomes_Raw[i] = weighted_median(votes_filled[:,i],
                                                           PlayerInfo["SmoothRep"].flatten())
 
         # .5 is obviously undesireable, this function travels from 0 to 1
         # with a minimum at .5
-        Certainty = abs(2 * (DecisionOutcomes_Raw - .5))
+        Certainty = abs(2 * (DecisionOutcomes_Raw - 0.5))
 
         # Grading Authors on a curve.
         ConReward = self.GetWeight(Certainty)
@@ -403,7 +371,7 @@ class Oracle(object):
         # Discriminate Based on Contract Type
         DecisionOutcome_Adj = []
         for i, raw in enumerate(DecisionOutcomes_Raw):
-            DecisionOutcome_Adj.append(self.Catch(raw, CatchP))
+            DecisionOutcome_Adj.append(self.Catch(raw))
             if ScaledIndex[i]:
                 DecisionOutcome_Adj[i] = raw
 
@@ -411,11 +379,11 @@ class Oracle(object):
         for i, raw in enumerate(DecisionOutcomes_Raw):
             DecisionOutcome_Final.append(DecisionOutcome_Adj[i])
             if ScaledIndex[i]:
-                DecisionOutcome_Final[i] *= (Scales[i]["max"] - Scales[i]["min"])
+                DecisionOutcome_Final[i] *= (self.decision_bounds[i]["max"] - self.decision_bounds[i]["min"])
 
         # Participation
         # Information about missing values
-        NAmat = M0 * 0
+        NAmat = self.votes * 0
         NAmat[NAmat.mask] = 1  # indicator matrix for missing
 
         # Participation Within Decisions (Columns)
@@ -431,7 +399,7 @@ class Oracle(object):
         PercentNA = 1 - np.mean(ParticipationC)
 
         # Possibly integrate two functions of participation? Chicken and egg problem...
-        if Verbose:
+        if self.verbose:
             print('*Participation Information*')
             print('Voter Turnout by question')
             print(ParticipationC)
@@ -448,8 +416,8 @@ class Oracle(object):
         ColBonus = NAbonusC * PercentNA + ConReward * (1 - PercentNA)
 
         Output = {
-            'Original': M0.base,
-            'Filled': Filled.base,
+            'Original': self.votes.base,
+            'Filled': votes_filled.base,
             'Agents': {
                 'OldRep': PlayerInfo['OldRep'][0],
                 'ThisRep': PlayerInfo['ThisRep'][0],
